unit NSock;

interface

uses Windows, Messages, WinSock;

type
  TNSock = class(TObject)
  private
    FSoc: TSocket;
    FHostEnt: array[0..MAXGETHOSTSTRUCT-1] of Char;
  protected
    procedure CreateSocket(AType: Integer);
  public
    ResultCode: Integer;
    OnRecv: procedure (Sender: TObject) of object;
    OnGetHostByName: procedure (const IpAddr: string) of object;
    OnConnect: procedure (Sender: TObject) of object;
    OnClose: procedure (Sender: TObject) of object;
    OnError: procedure (Sender: TObject; ErrorCode: Integer) of object;
    destructor Destroy; override;
    function Recv(BufSize: Integer): string;
    procedure GetHostByName(const HostName: string);
    procedure Close;
  end;

  TNUdp = class(TNSock)
  public
    constructor Create(LocalPort: Integer);
    function Send(Data: string; const IpAddr: string; Port: Integer): Integer;
  end;

  TNTcp = class(TNSock)
  public
    procedure Connect(const IpAddr: string; Port: Integer);
    function Send(Data: string): Integer;
  end;

function GetHostName: string;

implementation

const
  WSOCK_GETHOST = WM_USER + 10;
  WSOCK_SELECT = WM_USER + 11;
  MAXOBJECT = 256;

type
  TObjMapItem = record
    Key: Integer;
    Obj: TObject;
  end;

  TObjMap = class(TObject)
    FItems: array[0..MAXOBJECT-1] of TObjMapItem;
    function Get(Key: Integer): TObject;
    procedure Put(Key: Integer; Obj: TObject);
    procedure Del(Obj: TObject);
    property Items[Key: Integer]: TObject read Get write Put; default;
  end;

var
  WSAData: TWSAData;
  Wnd: HWND;
  SocMap: TObjMap;
  GetHostMap: TObjMap;

function TObjMap.Get(Key: Integer): TObject;
var
  i: Integer;
begin
  for i := 0 to High(FItems) do
    if FItems[i].Key = Key then
    begin
      Result := FItems[i].Obj;
      Exit;
    end;
  Result := nil;
end;

procedure TObjMap.Put(Key: Integer; Obj: TObject);
var
  i: Integer;
begin
  for i := 0 to High(FItems) do
    if FItems[i].Obj = nil then
    begin
      FItems[i].Key := Key;
      FItems[i].Obj := Obj;
      Exit;
    end;
end;

procedure TObjMap.Del(Obj: TObject);
var
  i: Integer;
begin
  for i := 0 to High(FItems) do
    if FItems[i].Obj = Obj then
    begin
      FItems[i].Key := 0;
      FITems[i].Obj := nil;
    end;
end;

function WindowProc(hWnd: HWND; Msg: UINT; wParam: WPARAM; lParam: LPARAM): LRESULT; stdcall;
var
  Obj: TNSock;
begin
  Result := 0;
  case Msg of
    WSOCK_GETHOST:
      begin
        Obj := GetHostMap[wParam] as TNSock;
        if Assigned(Obj) then
        begin
         if Assigned(Obj.OnGetHostByName) then
           if WSAGETASYNCERROR(lParam) = 0 then
             Obj.OnGetHostByName(inet_ntoa(PInAddr(PHostEnt(@Obj.FHostEnt).h_addr_list^)^))
           else
             Obj.OnGetHostByName(''); // Error
        end;
        GetHostMap.Del(Obj);
      end;
    WSOCK_SELECT:
      begin
        Obj := SocMap[wParam] as TNSock;
        if Assigned(Obj) then
          if WSAGETSELECTERROR(lParam) = 0 then
            case WSAGETSELECTEVENT(lParam) of
              FD_READ:
                if Assigned(Obj.OnRecv) then
                  Obj.OnRecv(Obj);
              FD_CONNECT:
                if Assigned(Obj.OnConnect) then
                  Obj.OnConnect(Obj);
              FD_CLOSE:
                begin
                  if Assigned(Obj.OnRecv) then
                    Obj.OnRecv(Obj);
                  if Assigned(Obj.OnClose) then
                    Obj.OnClose(Obj);
                  Obj.Close;
                end;
            end
          else begin
            if Assigned(Obj.OnError) then
              Obj.OnError(Obj, WSAGETSELECTERROR(lParam));
            Obj.Close;
          end;
      end;
  else
    Result := DefWindowProc(hWnd, Msg, wParam, lParam);
  end;
end;

function MakeWindow: Boolean;
const
  ClassName = 'TNSockSockWindowClass';
var
  wc: TWndClass;
begin
  if not GetClassInfo(HInstance, ClassName, wc) then
  begin
    FillChar(wc, SizeOf(wc), 0);
    wc.lpfnWndProc := @WindowProc;
    wc.hInstance := HInstance;
    wc.lpszClassName := ClassName;
    if RegisterClass(wc) = 0 then
    begin
      Result := False;
      Exit;
    end;
  end;
  Wnd := CreateWindowEx(WS_EX_TOOLWINDOW , ClassName, nil, WS_POPUP,
                        0, 0, 64, 64, 0, 0, HInstance, nil);
  if Wnd = 0 then
  begin
    Result := False;
    Exit;
  end;
  Result := True;
end;

{ Functions }

function GetHostName: string;
const
  HOST_NAME_MAX = 255;
var
  buf: Array[0..HOST_NAME_MAX] of Char;
begin
  WinSock.GetHostName(buf, SizeOf(buf)); 
  Result := buf;
end;

{ TNSock }

destructor TNSock.Destroy;
begin
  Close;
end;

procedure TNSock.CreateSocket(AType: Integer);
begin
  FSoc := socket(AF_INET, AType, 0);
  if FSoc = INVALID_SOCKET then
  begin
    ResultCode := 1;
    Exit;
  end;
  if WSAAsyncSelect(FSoc, Wnd, WSOCK_SELECT, FD_READ or FD_CONNECT or FD_CLOSE) = SOCKET_ERROR then
  begin
    WinSock.closesocket(FSoc);
    ResultCode := 1;
    Exit;
  end;

  SocMap[FSoc] := Self;

  ResultCode := 0;
end;

procedure TNSock.Close;
begin
  if FSoc <> INVALID_SOCKET then
  begin
    shutdown(FSoc, 2);
    WinSock.closesocket(FSoc);
    FSoc := INVALID_SOCKET;
  end;
  SocMap.Del(Self);
end;

procedure TNSock.GetHostByName(const HostName: string);
var
  hGetHost: THandle;
begin
  if inet_addr(PChar(HostName)) = INADDR_NONE then
  begin
    hGetHost := WSAAsyncGetHostByName(Wnd, WSOCK_GETHOST, PChar(HostName),
                                      FHostEnt, SizeOf(FHostEnt));
    if hGetHost = 0 then
    begin
      if Assigned(OnGetHostByName) then
        OnGetHostByName(''); // Error
      Exit;
    end;

    GetHostMap[hGetHost] := Self;
  end
  else if Assigned(OnGetHostByName) then
    OnGetHostByName(HostName);
end;

function TNSock.Recv(BufSize: Integer): string;
var
  len: Integer;
begin
  SetString(Result, nil, BufSize);
  len := WinSock.recv(FSoc, Result[1], Length(Result), 0);

  if len = SOCKET_ERROR then
  begin
    ResultCode := 1;
    SetLength(Result, 0);
    Exit;
  end;

  SetLength(Result, len);
  ResultCode := 0;
end;


{ TNUdp }

constructor TNUdp.Create(LocalPort: Integer);
var
  saddr: TSockAddr;
begin
  inherited Create;
  CreateSocket(SOCK_DGRAM);
  if ResultCode = 0 then
  begin
    if LocalPort <> -1 then
    begin
      FillChar(saddr, SizeOf(saddr), 0);
      with saddr do
      begin
        sin_family := AF_INET;
        sin_port := htons(LocalPort);
        sin_addr.S_addr := 0;
      end;
      bind(FSoc, saddr, SizeOf(saddr));
      // TODO: check error
    end;
  end;
end;

function TNUdp.Send(Data: string; const IpAddr: string; Port: Integer): Integer;
var
  saddr: TSockAddr;
begin
  FillChar(saddr, SizeOf(saddr), 0);
  with saddr do
  begin
    sin_family := AF_INET;
    sin_port := htons(Port);
    sin_addr.s_addr := inet_addr(PChar(IpAddr));
  end;
  Result := sendto(FSoc, Data[1], Length(Data), 0, saddr, SizeOf(saddr));
end;


{ TNTcp }

procedure TNTcp.Connect(const IpAddr: string; Port: Integer);
var
  saddr: TSockAddr;
begin
  CreateSocket(SOCK_STREAM);
  // TODO: check ResultCode
  FillChar(saddr, SizeOf(saddr), 0);
  with saddr do
  begin
    sin_family := AF_INET;
    sin_port := htons(Port);
    sin_addr.s_addr := inet_addr(PChar(IpAddr));
  end;
  if WinSock.connect(FSoc, saddr, SizeOf(saddr)) = SOCKET_ERROR then
    if WSAGetLastError <> WSAEWOULDBLOCK then
    begin
      ResultCode := 1;
      Exit;
    end;
end;

function TNTcp.Send(Data: string): Integer;
begin
  Result := WinSock.send(FSoc, Data[1], Length(Data), 0);
end;

initialization
  if WSAStartUp($0101, WSAData) <> 0 then
  begin
    MessageBox(0, 'Winsock initilized error', nil, MB_OK);
    Halt(255);
  end;
  if not MakeWindow then
  begin
    MessageBox(0, 'Create NSock Window failed', nil, MB_OK);
    Halt(255);
  end;
  SocMap := TObjMap.Create;
  GetHostMap := TObjMap.Create;

finalization
  if Wnd <> 0 then
    DestroyWindow(Wnd);
  SocMap.Free;
  GetHostMap.Free;
  WSACleanUp;

end.
