unit Sntpc;

interface

uses Windows, WinSock, NSock;

type
  PNtpTs = ^TNtpTs;
  TNtpTs = packed record
    i: Integer;
    f: Integer;
  end;

  PNtpPacket = ^TNtpPacket;
  TNtpPacket = packed record
    li_vn_mode: Byte; // li:2 vn:3 mode:3
    stratum: Byte;
    poll: Byte;
    precision: Shortint;
    rootdelay: Integer;
    rootdispersion: Integer;
    refid: Integer;
    reftime: TNtpTs;
    originate: TNtpTs;
    receive: TNtpTs;
    transmit: TNtpTs;
  end;

  TSntpcSyncInfo = record
    IpAddr: string;
    Stratum: Integer;
    Offset: Double;
    Delay: Double;
    Adjusted: Boolean;
  end;

  TSntpc = class(TObject)
  private
    FUdp: TNUdp;
    FOriginate: Double;
    FSyncing: Boolean;
    FAdjustOffset: Double;
    FIpAddr: string;
    procedure Recv(Sender: TObject);
    procedure GetHostByName(const IpAddr: string);
    function IsValidPkt(const pkt: TNtpPacket): Boolean;
    procedure Error(Code: Integer);
    function NtpTsToD(const ts: TNtpTs): Double;
    function GetSysTimeD: Double;
    procedure SetSysTimeD(Value: Double);
  public
    Port: Integer;
    OnSend: procedure(const IpAddr: string) of object;
    OnSynced: procedure(const SyncInfo: TSntpcSyncInfo) of object;
    OnError: procedure(Code: Integer) of object;
    constructor Create;
    destructor Destroy; override;
    procedure Sync(const Server: string);
    procedure Abort;
    property Syncing: Boolean read FSyncing;
    property AdjustOffset: Double read FAdjustOffset write FAdjustOffset ;
  end;

const
  SNTPC_ERR_SYNC = 1;
  SNTPC_ERR_SEND = 2;
  SNTPC_ERR_HOSTNOTFOUND = 3;

implementation

constructor TSntpc.Create;
begin
  Port := 123;
  AdjustOffset := 0.5;  
end;

destructor TSntpc.Destroy;
begin
  if Assigned(FUdp) then
    FUdp.Free;
end;

procedure TSntpc.Error(code: Integer);
begin
  FSyncing := False;
  if Assigned(OnError) then
    OnError(code);
end;

function TSntpc.IsValidPkt(const pkt: TNtpPacket): Boolean;
begin
  Result := True;
  with pkt do
  begin
    if (li_vn_mode shr 6) > 2 then // li
      Result := False
    else if (li_vn_mode shr 3 and 7) <> 4 then // vn
      Result := False
    else if (li_vn_mode and 7) <> 4 then // mode
      Result := False
    else if stratum > 14 then
      Result := False
    else if (transmit.i = 0) and (transmit.f = 0) then
      Result := False;
  end;
end;

procedure TSntpc.Recv(Sender: TObject);
var
  pkt: TNtpPacket;
  s: string;
  rec, xmit, delay, ofs: Double;
  adjusted: Boolean;
  syncinfo: TSntpcSyncInfo;
  Dest: Double;
begin
  Dest := GetSysTimeD;
  s := FUdp.Recv(SizeOf(pkt));
  FUdp.Free;
  FUdp := nil;

  if Length(s) <> SizeOf(pkt) then
  begin
    Error(SNTPC_ERR_SYNC);
    Exit;
  end;

  Move(s[1], pkt, SizeOf(pkt));
  if not IsValidPkt(pkt) then
  begin
    Error(SNTPC_ERR_SYNC);
    Exit;
  end;

  // Do sync
  rec := NtpTsToD(pkt.receive);
  xmit := NtpTsToD(pkt.transmit);
  delay := (Dest - FOriginate) - (rec - xmit);
  ofs := ((rec - FOriginate) + (xmit - Dest)) / 2;
  adjusted := (ofs > FAdjustOffset) or (ofs < -FAdjustOffset);
  if adjusted then
    SetSysTimeD(GetSysTimeD + ofs);

  FSyncing := False;

  if Assigned(OnSynced) then
  begin
    syncinfo.IpAddr := FIpAddr;
    syncinfo.Stratum := pkt.stratum;
    syncinfo.Offset := ofs;
    syncinfo.Delay := delay;
    syncinfo.Adjusted := adjusted;
    OnSynced(syncinfo);
  end;
end;

procedure TSntpc.GetHostByName(const IpAddr: string);
var
  s: string;
begin
  if IpAddr = '' then
  begin
    Error(SNTPC_ERR_HOSTNOTFOUND);
    Exit;
  end;
  FIpAddr := IpAddr;
  SetString(s, nil, SizeOf(TNTPPacket));
  FillChar(s[1], SizeOf(TNTPPacket), 0);
  PNtpPacket(@s[1]).li_vn_mode := $23; // Version Number = 4, Mode = 3 (client);
  if Assigned(OnSend) then
    OnSend(FIpAddr);
  FOriginate := GetSysTimeD;
  if FUdp.Send(s, FIpAddr, Port) = SOCKET_ERROR then
    Error(SNTPC_ERR_SEND);
end;

function TSntpc.NtpTsToD(const ts: TNtpTs): Double;
const
  dUInt32Max = 4294967295.0;
var
  Sec, Frac: Double;
begin
  Sec := ntohl(ts.i);
  if Sec < 0 then
    Sec := Sec + dUInt32Max + 1;
  Frac := ntohl(ts.f);
  if Frac < 0 then
    Frac := Frac + dUInt32Max + 1;
  Result := Sec + Frac / (dUInt32Max+1);
end;

function TSntpc.GetSysTimeD: Double;
var
  st: TSystemTime;
  d: Integer;
begin
  GetSystemTime(st);
  with st do
  begin
    if wMonth <= 2 then
    begin
      Inc(wMonth, 12);
      Dec(wYear);
    end;
    // days from 1900/01/01
    d := (wYear*365)+(wYear div 4)-(wYear div 100)+(wYear div 400) +
         (wMonth*30) + (((wMonth+1)*3) div 5) + wDay - 33 - 693960;
    Result := ((d-1) * 86400.0 + wHour * 3600.0 + wMinute * 60.0 + wSecond)
              + wMilliseconds / 1000.0;
  end;
end;

procedure TSntpc.SetSysTimeD(Value: Double);
var
  ft: TFileTime;
  st: TSystemTime;
begin
  // TODO: FileTimegȂ悤ɂ
  st.wYear := 1900;
  st.wMonth := 1;
  st.wDay := 1;
  st.wHour := 0;
  st.wMinute := 0;
  st.wSecond := 0;
  st.wMilliseconds := 0;
  SystemTimeToFileTime(st, ft);
  Comp(ft) := Comp(ft) + Value * 10000000;
  FileTimeToSystemTime(ft, st);
  SetSystemTime(st);
end;

procedure TSntpc.Sync(const Server: string);
begin
  if FSyncing then
    Exit
  else
    FSyncing := True;
  if Assigned(FUdp) then
    FUdp.Free;
  FUdp := TNUdp.Create(-1);
  FUdp.OnRecv := Recv;
  FUdp.OnGetHostByName := GetHostByName;
  FUdp.GetHostByName(Server);
end;

procedure TSntpc.Abort;
begin
  if FSyncing then
  begin
    FUdp.Free;
    FUdp := nil;
    FSyncing := False;
  end;
end;

end.

